/* $Workfile: T2pT1RobBit.c $       $Revision: 53 $      $Modtime: 9/18/08 3:01p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T2pT1RobBit.c

    A simple example console application demonstrating the use of OTX Driver
    to access T1 Rob-Bit signalling bits with with following boards:

      Thor-2(8)-PCI

    or OTX PCI Plus series boards: Thor-2(4,8)-PCI-Plus 

    or OTX PCI Express series boards: Thor-2-PCI-Express

    or Thor-2-PCMCIA series boards: Thor-2-PCMCIA(-Plus,-Ex,-Pro,-CST)

    or Sleipnir-1-PCI-Plus

    or OTX ExpressCard series board: Thor-2(4)-ExpressCard

    or Alvis-4-CSI or Alvis-4M-CSI

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2 T1 Lines (8 for Thor-8-PCI) and expects them to be looped 
    with an external loop cable. The T1 lines are used for accessing Bit Robbed 
    Channel Associated Signaling information.

    Channel 1 through 23 are used CAS information (referred to as CAS channels).
    Channel 24 is used the HDLC framing, so this timeslot is configured as
    a clear-channel.

    Keystroke commands are used to select the current CAS channel, set a ABCD-bit
    value on a certain CAS channel, and read the ABCD-bit values on all CAS channels.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"


#define DEMO_DEFAULT_LOG_FILE "T2pT1RobBit.log"
#define DEMO_DEFAULT_BOARD_NO 0
#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call

typedef struct DemoEnvT {
  char           *pszInputStr;
  OTX_BOOL        bExitOnError;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  OTX_INT32       nBoardType;
  int             coLi;
  int             boardNo;
  OTX_BOOL        bChannelizedAbcdBitAccess; // If OTX_TRUE then report ABCD data per channel in the OTX_T1E1_EC_BITROB_SIGNALING event
} DemoEnvT;

/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OtxT1RobBitDataS  g_sqSendRobBitData[8];
OTX_UINT8   g_sqHdlcFrame[OD_HDLC_MAX_TX_HDLC_FRAME_LEN];
OTX_INT32   g_nSleepTimeBetweenHdlcFrames = 1000; // in milliseconds
OTX_INT32   g_nSleepTimeBetweenBitrobEvents = 1000; // in milliseconds
OTX_INT32   g_cbHdlcFrame = 10;
OTX_INT32   g_coHdlcSenderFrames = 1;
OTX_BOOL    g_bSimulateBitRobSignalling = OTX_FALSE;
OTX_BOOL    g_bSimulateHdlcSignalling = OTX_FALSE;
#if 0 // for testing a full event queue condition
OTX_BOOL    g_bSimulateBusy = OTX_FALSE;
#endif
OTX_BOOL    g_bBitRobSignallingVerbose = OTX_TRUE;
OTX_BOOL    g_bHdlcSignallingVerbose = OTX_TRUE;
OTX_INT32   g_coBitRobSignallingEvents = 0;

struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[2];        // Used for Thor-4-PCI-Plus and Thor-8-PCI(-Plus)
  OTX_HANDLE hLi[8];
} g_hPhysDevs;


struct {
  OTX_HANDLE hLiIn[8];
  OTX_HANDLE hLiOut[8];
} g_hPipes;


struct {
  OTX_HANDLE hT1E1Cas[8];

  // The HDLC Receiver/Sender is connected to TS24 on the T1 span.
  OTX_HANDLE hHdlcSender[8];
  OTX_HANDLE hHdlcReceiver[8];
} g_hLogicDevs;


struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;    // D16V86 apps cannot use event objects
    result = OTX_S_OK;
  }

  DemoCheckRc(result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(/*128*/16, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; // Only used for Thor-4-PCI-Plus, Thor-8-PCI(-Plus)

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening Thor-2-PCI board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf("Opening Thor-2-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    DemoLogf("Opening Thor-2-PCI-Express board\n");
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    DemoLogf("Opening Thor-4-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf("Opening Thor-8-PCI board\n");
    break;
  case OTX_DEVICE_THOR_PCMCIA:
    DemoLogf("Opening Thor-2-PCMCIA variant board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    break;
  case OTX_DEVICE_SLEIPNIR_1_PCI:
    DemoLogf("Opening Sleipnir-1-PCI board\n");
    break;
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    DemoLogf("Opening Thor-2-ExpressCard board\n");
    break;
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
    DemoLogf("Opening Thor-4-ExpressCard board\n");
    break;
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
    DemoLogf("Opening Alvis-4-CSI or Alvis-4M-CSI board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(result);

  // Get handles to T1/E1 line interface devices 
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_SLEIPNIR_1_PCI:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    // On these boards the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
    {
      // On the Thor-4-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. 
      for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
      }
    }
    break;
  case OTX_DEVICE_THOR_8_PCI:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    // On the Thor-8-PCI(-Plus) board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
    }

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    // On the Thor-2-ExpressCard board the E1T1 devices are hosted by a dual E1T1 device. 
    DemoLogf("Open OTX_DEVICE_DUAL_T1E1\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DUAL_T1E1, 0, 0, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[0]); DemoCheckRc(result);

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_DUAD_T1E1)\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[0], OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
 
    // On the Thor-4-ExpressCard board the E1T1 devices are hosted by a Qual E1T1 device. 
    DemoLogf("Open OTX_DEVICE_QUAD_T1E1\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, 0, 0, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[0]); DemoCheckRc(result);

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1)\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[0], OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }

    // Thor-4-ExpressCard by default is set to receive on the four lines 
    // (monitor mode) with all transmitters disabled (tristate). This 
    // demo however needs the card to support the transmit data so the 
    // card will use 2 receivers and the 2 transmitters.
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hThorController0, 
                                         OTX_ATTR_BRD_MONITOR_ONLY, 
                                         OTX_FALSE); 
    DemoCheckRc(result);
    DemoLogf("Configure Thor-4-ExpressCard for data transmit and receive\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  

}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(result);

    // Since the demo application assumes that Li0 is looped to Li1, enable Automatic CRC bit inversion for one of the Line Intefaces (not both)
    if (i == 0) {
      g_T1E1CfgData.t1CfgData.bAutoCrcBitInversion = OTX_FALSE;
    } else {
      g_T1E1CfgData.t1CfgData.bAutoCrcBitInversion = OTX_TRUE;
    }

    result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(result);

    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_T1); DemoCheckRc(result);

    if (g_demoEnv.bChannelizedAbcdBitAccess) {
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_CHANNELIZED_ABCD_BIT_ACCESS, OTX_TRUE); DemoCheckRc(result);
    }

  }

}



/*******************************************************************************
  @func
    Create the needed Logical Devices
  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.
*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create CAS Sender and Receiver logical devices hosted in the physical Line Interface devices
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hLi[i], OTX_LDEVICE_T1_ROB_BIT_ACCESS, g_hEventQueue, i, &g_hLogicDevs.hT1E1Cas[i]); DemoCheckRc(result);
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create Hdlc Senders and Receivers logical devices hosted in the physical Line Interface devices
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hLi[i], OTX_LDEVICE_HDLC_SENDER, g_hEventQueue, i, &g_hLogicDevs.hHdlcSender[i]); DemoCheckRc(result);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hLi[i], OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, i, &g_hLogicDevs.hHdlcReceiver[i]); DemoCheckRc(result);
  }

}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 ts;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create an incoming and outgoing Pipe representing the external T1/E1
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiIn[i]); DemoCheckRc(result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiOut[i]); DemoCheckRc(result);

    // Connect external T1/E1 pipes. Only connect one end of each pipe. External Interface is LI Highway Number 1
    ts = 24;
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[ts] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hLiIn[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(result);
    result = OtxDrvConnectPipe(g_hPipes.hLiOut[i], OTX_PIPE_INPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(result);

    // Connect HDLC Sender and Receiver to the external T1/E1 Pipes
    result = OtxDrvConnectLogicalDevice(g_hPipes.hLiIn[i], -1, g_hLogicDevs.hHdlcReceiver[i], -1); DemoCheckRc(result);
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hHdlcSender[i], -1, g_hPipes.hLiOut[i], -1); DemoCheckRc(result);
  }
}


/*******************************************************************************
  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.
*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32 i;

  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Initialize the Hdlc Sender and Receiver
    result = OtxHdlcSenderInit(
               g_hLogicDevs.hHdlcSender[i], // Handle to logical HDLC Filter Device
               OTX_HDLC_FFT_FLAGS,          // Fill character(s) to be send between HDLC frames
               2,                           // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
               OTX_HDLC_CRC16,              // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
               OTX_FALSE,                   // If OTX_TRUE then all bits are inverted when transmitted and before received
               OTX_TRUE,                    // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
               OTX_HDLC_TR_NA);             // HDLC Transmission Rate for V.110 and X.30 modes
               DemoCheckRc(result);

    result = OtxHdlcReceiverInit(
               g_hLogicDevs.hHdlcReceiver[i], // Handle to logical HDLC Filter Device
               OTX_HDLC_FFT_ALL_ONES,          // Fill character(s) to be send between HDLC frames
               1,                           // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
               OTX_HDLC_CRC16,              // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
               OTX_FALSE,                   // If OTX_TRUE then all bits are inverted when transmitted and before received
               OTX_TRUE,                    // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
               OTX_HDLC_TR_NA,              // HDLC Transmission Rate for V.110 and X.30 modes
               OTX_TRUE);                   // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
               DemoCheckRc(result);

  }
}


/*******************************************************************************

  @func
    Enable Configured Devices

*******************************************************************************/
void DemoEnableDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1;

  // Enable Physical Devices
  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

  // If board type is Thor-4-PCI-Plus or Thor-2(4)-ExpressCard, then enable Quad E1T1 device
  if ( (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) || (g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_4_CSI) ) {
      
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
  }

  // If board type is Thor-8-PCI(-Plus) then enable Quad E1T1 device
  if ( (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) ) {
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
  }


  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

  // Enable Logical T1/E1 Cas Devices
  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hT1E1Cas[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

  // Enable Logical HDLC Devices 
  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hHdlcSender[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    result = OtxDrvEnable(g_hLogicDevs.hHdlcReceiver[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }
}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CAS_BR;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_AMI;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_AMI;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_TRUE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0x800000;  // 0x800000 means that channel 24 (1-count) is a clear channel. E.g. used for 64 kbps signalling

  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;
}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
#if 1
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
#else  // For testing a full event queue condition
    {
      OTX_UINT32 i;
      OTX_UINT32 coEvents;
      DemoLogf("Event signalled. Reading one event\n");
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(result);
      if (OTX_RESULT_CODE(result) != OTX_S_OK) {
        DemoLogf("OtxDrvGetEventData() failed to return an event [1]\n");
        continue;
      } else {
        DemoPrintEvent(&eventData);
      }
      
      result = OtxDrvGetAttributeValueUint32(g_hEventQueue, OTX_ATTR_QUEUE_USAGE, &coEvents); DemoCheckRc(result);
      for (i = 0; i < coEvents; i++) {
        DemoLogf("Reading event %u of %u\n", i+1, coEvents);

        while (g_bSimulateBusy) {
          Sleep(500);
        }
        result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(result);
        if (OTX_RESULT_CODE(result) != OTX_S_OK) {
          DemoLogf("OtxDrvGetEventData() failed to return an event [2]\n");
          break;
        } else {
          DemoPrintEvent(&eventData);
        }
      }
      DemoLogf("Sleeping until next event notification\n");
    }
#endif
  }

  return 0;

}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}


/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  if (g_bBitRobSignallingVerbose) {
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
  }

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_LI_T1E1:
    // Event from T1E1 physical device (Li)
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      DemoLogf("E1/T1[%d] No string for event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    if (pEventData->m_nParam) {
      DemoLogf("E1/T1[%d] Event: %s (0x%x)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
    } else {
      DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
    }

    break;


  case OTX_LDEVICE_T1_ROB_BIT_ACCESS:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if ((pEventData->m_nCode == OTX_T1E1_EC_BITROB_SIGNALING) && g_demoEnv.bChannelizedAbcdBitAccess) {
        g_coBitRobSignallingEvents++;
        if (g_bBitRobSignallingVerbose) {
          DemoLogf("BitRob Device[%d] Event: %s (Channel:%.2d ABCD:0x%X)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam >> 8, pEventData->m_nParam & 0x0F);
        }
        
        // Print the counter every so often
        if ((g_coBitRobSignallingEvents % 500) == 0) {
          DemoLogf("g_coBitRobSignallingEvents = %d\n", g_coBitRobSignallingEvents);
        }
      } else {
        
        if (pEventData->m_nParam) {
          DemoLogf("BitRob Device[%d] Event: %s (0x%x)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
        } else {
          DemoLogf("BitRob Device[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
        }
      }
    } else {
      DemoLogf("BitRob Device[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;


  case OTX_LDEVICE_HDLC_RECEIVER:
    // Event from HDLC Receiver 
    if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
      OTX_UINT32 i;
      OTX_UINT8  buf[1024];
      OTX_UINT32 nFrameLen;
      OTX_UINT32 nFrameSeq;
      OtxHdlcFrameEndCodeE nFrameEndCode;
      OTX_DATETIME nTimeStamp;
      
      otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
      if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
                      
        if (g_bHdlcSignallingVerbose) {
          DemoLogf("HDLC Frame Received, Sequence Number = %d\n", pEventData->m_nParam);
          
          DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
          for (i = 0; i < nFrameLen && i < 1024; i++) {
            DemoLogf("%.2X ", buf[i]);
          }
          DemoLogf("\n");
        } else {
          if ((pEventData->m_nParam % 500) == 0) {
            DemoLogf("HDLC Frame Received, Sequence Number = %d\n", pEventData->m_nParam);
          }
        }
        
      } else {
        DemoLogf("HdlcReceiver[%d] OtxHdlcReadData() was unsuccessful.\n", pEventData->m_nSourceId);
        DemoCheckRc(otxRc);
      }
      
      
    } else {
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;


  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}


/*******************************************************************************
  @func
    Extract a specific channel ABCD bit value from the OtxT1RobBitDataS struct.
*******************************************************************************/
char *DemoGetCasChannelValue(
  IN OTX_INT32 nCasCh,
  IN OtxT1RobBitDataS  *pRobBitData,
  OUT char *pszValue  // @parm ABCD bits (in that order) as a string
)
{
  pszValue[0] = '\0';
  
  if (nCasCh < 1 || nCasCh > 24) {
    DemoLogf("Error: CAS Channel out of range (%d)\n", nCasCh);
    return pszValue;
  }

  sprintf(pszValue + strlen(pszValue), "%d",  (pRobBitData->chA >> (nCasCh - 1)) & 0x01 ? 1 : 0);
  sprintf(pszValue + strlen(pszValue), "%d",  (pRobBitData->chB >> (nCasCh - 1)) & 0x01 ? 1 : 0);
  sprintf(pszValue + strlen(pszValue), "%d",  (pRobBitData->chC >> (nCasCh - 1)) & 0x01 ? 1 : 0);
  sprintf(pszValue + strlen(pszValue), "%d",  (pRobBitData->chD >> (nCasCh - 1)) & 0x01 ? 1 : 0);

  return pszValue;
}


/*******************************************************************************
  @func
    Sets a specific CAS-channel ABCD bit value in the OtxT1RobBitDataS struct.
*******************************************************************************/
char *DemoSetCasChannelValue(
  IN OTX_INT32 nCasCh,
  IN OtxT1RobBitDataS *pRobBitData,
  IN char *pszValue  // @parm ABCD bits (in that order) as a string
)
{
  OTX_INT32 i;
  OTX_UINT32 *pChX = &(pRobBitData->chA);
  
  if (nCasCh < 1 || nCasCh > 24) {
    DemoLogf("Error: CAS Channel out of range (%d)\n", nCasCh);
    return pszValue;
  }

  for (i = 0; i < 4; i++) {
    switch (i) {
    case 0:
      pChX = &(pRobBitData->chA);
      break;
    case 1:
      pChX = &(pRobBitData->chB);
      break;
    case 2:
      pChX = &(pRobBitData->chC);
      break;
    case 3:
      pChX = &(pRobBitData->chD);
      break;
    }

    switch (pszValue[i]) {
    case '1':
      *pChX |=   0x01 << (nCasCh - 1);
      break;
    case '0':
      *pChX &= ~(0x01 << (nCasCh - 1));
      break;
    default:
      DemoLogf("Error: Invalid bit value (%c). Expected '0' or '1'\n", pszValue[i]);
      return pszValue;
    }
  }

  return pszValue;
}
  


/*******************************************************************************
  @func
    Prints the Rob Bit data for all 24 CAS channels
*******************************************************************************/
void DemoPrintAllCasChannelsData(
  OTX_INT32 nCasDev
)
{
  OTX_RESULT result;

  DemoLogf("Reading received bitrob data on LI#%d ... \n", nCasDev);

  if (g_demoEnv.bChannelizedAbcdBitAccess) {
    OTX_INT32 i;
    OTX_UINT8 nABCDBits;

    for (i = 0; i < 24; i++) {
      result = OtxT1GetChannelizedRobBitData(g_hLogicDevs.hT1E1Cas[nCasDev], i+1, &nABCDBits); DemoCheckRc(result);
      DemoLogf("   Channel %.2d: 0x%X", i+1, nABCDBits);
      if (i % 2) {
        DemoLogf("\n");
      }
    }

  } else {
    OTX_INT32 i;
    OtxT1RobBitDataS  rcvRobBitData;
    char szABCDbuf[4+1];  // +1 for the '\0'
    
    result = OtxT1GetRobBitData(g_hLogicDevs.hT1E1Cas[nCasDev], &rcvRobBitData); DemoCheckRc(result);
    if (OTX_RESULT_CODE(result) != OTX_S_OK) {
      return;
    }
    DemoLogf("A = %.6X, B = %.6X, C = %.6X, D = %.6X\n", rcvRobBitData.chA, rcvRobBitData.chB, rcvRobBitData.chC, rcvRobBitData.chD);
    DemoLogf("\n");
    DemoLogf("            ABCD              ABCD\n");
    DemoLogf("            ====              ====\n");
    for (i = 0; i < 24; i++) {
      DemoLogf("Channel %.2d: ", i+1);
#if 0
      DemoLogf("%d",  (rcvRobBitData.chA >> i) & 0x01 ? 1 : 0);
      DemoLogf("%d",  (rcvRobBitData.chB >> i) & 0x01 ? 1 : 0);
      DemoLogf("%d",  (rcvRobBitData.chC >> i) & 0x01 ? 1 : 0);
      DemoLogf("%d",  (rcvRobBitData.chD >> i) & 0x01 ? 1 : 0);
#else
      DemoLogf("%s", DemoGetCasChannelValue(i+1, &rcvRobBitData, szABCDbuf));
#endif
      DemoLogf("  ");
      if (i % 2) {
        DemoLogf("\n");
      }
    }
    DemoLogf("\n");
  } 
}


/*******************************************************************************
  @func
    Send random RBS data on random channels
*******************************************************************************/
void *DemoSimulateBitRobSignalling(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_DATETIME nTimeStamp;

  /* Seed the random-number generator with current time so that
   * the numbers will be different every time we run.
   */
  OtxDrvGetCurrentTime(&nTimeStamp);
  srand(nTimeStamp.sUint32.nLow);

  while (g_bSimulateBitRobSignalling) {

    if (g_demoEnv.bChannelizedAbcdBitAccess) {
      OTX_UINT8 nABCDBits;
      OTX_INT32 nLi;
      OTX_INT32 nCasChannel;
      
      nLi = rand() % g_demoEnv.coLi;
      
      nCasChannel = (rand() % 24) + 1; // nCasChannel is a number between 1 and 24 (inclusive)
      
      nABCDBits = (OTX_UINT8)(rand() % 16); // 0x0 - 0xF
      
      if (g_bBitRobSignallingVerbose) {
        DemoLogf("Setting ABCD-bit data for CAS channel %d on Li%d (ABCD=0x%X)\n", nCasChannel, nLi, nABCDBits);
      }
      result = OtxT1SetChannelizedRobBitData(g_hLogicDevs.hT1E1Cas[nLi], nCasChannel, nABCDBits); //DemoCheckRc(result);
      if (OTX_RESULT_CODE(result) != OTX_S_OK) {
        DemoLogf("Setting ABCD-bit data for CAS channel %d on Li%d (ABCD=0x%X) failed.\n (result=%X)", nCasChannel, nLi, nABCDBits, result);
      }
      
    } else {
      
      g_bSimulateBitRobSignalling = OTX_FALSE;
      
      DemoLogf("Simulate BitRob Signalling is only supported in ChannelizedAbcdBitAccess mode. Thread aborting.");
    }


    OsaThreadSleep(g_nSleepTimeBetweenBitrobEvents);
  }

  DemoLogf("DemoSimulateBitRobSignalling() thread is terminating.\n");
  return 0;

}

/*******************************************************************************
  @func
    Send HDLC frames randomly
*******************************************************************************/
void *DemoSimulateHdlcSignalling(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_UINT8 sqHdlcFrame[128];
  OTX_DATETIME nTimeStamp;

  /* Seed the random-number generator with current time so that
   * the numbers will be different every time we run.
   */
  OtxDrvGetCurrentTime(&nTimeStamp);
  srand(nTimeStamp.sUint32.nLow);

  while (g_bSimulateHdlcSignalling) {
      OTX_INT32 nLi;
      OTX_INT32 cbFrameLen;
      OTX_INT32 i;
      
      nLi = rand() % g_demoEnv.coLi;

      cbFrameLen = (rand() % 30) + 3;

      for (i = 0; i < cbFrameLen; i++) {
        sqHdlcFrame[i] = (OTX_UINT8) i;
      }

      //DemoLogf("Sending HDLC frame from Li #%d (len=%d) ...", nLi, cbFrameLen);

      // Send HDLC Frame(s)
        result = OtxHdlcSendData(g_hLogicDevs.hHdlcSender[nLi], sqHdlcFrame, cbFrameLen, DEMO_TASK_REF, OTX_TIME_INFINITY); 
        if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
          // Driver IO Request buffer is full. Skip this frame.
        } else {
          DemoCheckRc(result);
        }


    OsaThreadSleep(g_nSleepTimeBetweenHdlcFrames);
  }

  DemoLogf("DemoSimulateHdlcSignalling() thread is terminating.\n");
  return 0;

}


/*******************************************************************************
  @func
    Parse the program's command line arguments and calls upon
    requested actions.

    There are some optional switches that can be used following
    the driver name on the command line. Switches are are
    separated by spaces, and can appear in arbitrary order.
    Switches consist of a hyphen and a letter:
*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr               = NULL;
  envP->bExitOnError              = OTX_FALSE;
  envP->boardNo                   = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType                = OTX_DEVICE_THOR_PCI;
  envP->coLi                      = 2;
  envP->bChannelizedAbcdBitAccess = OTX_TRUE;

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'U') {   // Un-Channelized ABCD bit access
          envP->bChannelizedAbcdBitAccess = OTX_FALSE;
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '2') {
            envP->nBoardType = OTX_DEVICE_THOR_PCI;
            envP->coLi = 2;
          } else if ( *(argv[i]+2) == 'T') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_PLUS;
            envP->coLi = 2;
          } else if ( *(argv[i]+2) == 'E') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_EXPRESS;
            envP->coLi = 2;
          } else if ( *(argv[i]+2) == '4') {
            envP->nBoardType  = OTX_DEVICE_THOR_4_PCI_PLUS;
            envP->coLi = 4;
          } else if ( *(argv[i]+2) == '8') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI;
            envP->coLi = 8;
          } else if ( *(argv[i]+2) == 'M') {
            envP->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
            envP->coLi = 2;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coLi = 8;
          } else if ( *(argv[i]+2) == 'S') {
            envP->nBoardType  = OTX_DEVICE_SLEIPNIR_1_PCI;
            envP->coLi = 1;
          } else if ( *(argv[i]+2) == 'C') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_EXPRESSCARD;
            envP->coLi = 2;
          } else if ( *(argv[i]+2) == 'R') {
            envP->nBoardType  = OTX_DEVICE_THOR_4_EXPRESSCARD;
            envP->coLi = 2; // use '2' in this demo 
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }
}


/*******************************************************************************
  @func
    Displays a help for the command line.
*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /T[2|T|E|4|8||M|P|S|C]: Board Type used for the demo, where\n");
  DemoLogf("                    2 = Use Thor-2-PCI Board (default)\n");
  DemoLogf("                    T = Use Thor-2-PCI-Plus Board\n");
  DemoLogf("                    E = Use Thor-2-PCI-Express Board\n");
  DemoLogf("                    4 = Use Thor-4-PCI-Plus Board\n");
  DemoLogf("                    8 = Use Thor-8-PCI Board\n");
  DemoLogf("                    M = Use Thor-2-PCMCIA Board\n");
  DemoLogf("                    P = Use Thor-8-PCI-Plus Board\n");
  DemoLogf("                    S = Use Sleipnir-1-PCI Board\n");
  DemoLogf("                    C = Use Thor-2-ExpressCard Board\n");
  DemoLogf(" /U           : Use un-channelized ABCD bit access\n");
  DemoLogf(" /Z<str>      : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************
  @func
    Displays a message on the screen and prints it to the logfile.
*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


static void DemoCheckRc(IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {
    OTX_RESULT otxRc;
    char msgStr[256];
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), msgStr);
    
    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}



/*******************************************************************************
  @func
    Processes User Commands
*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32 nCurrentCasChannel = 1;
  static OTX_INT32 nCurrentLi = 0;
  OTX_RESULT result;

  DemoLogf("Command '%c' pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-7 Select current Li. (currently Li=%d selected)\n", nCurrentLi);
    DemoLogf(" n   Increment current CAS channel (currently channel %d selected\n", nCurrentCasChannel);
    DemoLogf(" N   Decrement current CAS channel (currently channel %d selected\n", nCurrentCasChannel);
    DemoLogf(" s   Set transmitted ABCD-bits for current CAS channel on current Li\n");
    DemoLogf(" z   Zero transmitted ABCD-bits on current Li (all CAS channels)\n");
    DemoLogf(" r   Read the value of the received ABCD-bits on the other Li (all CAS channels)\n");
    DemoLogf(" R   Read the value of the received ABCD-bits on the current Li (all CAS channels)\n");
    DemoLogf(" l   Query Line Status of all Lis\n");
    DemoLogf(" t   Transmit HDLC frame on current Li (TS24)\n");
    DemoLogf(" y/Y Start/Stop Automatic simulation of bitrob signalling\n");
    DemoLogf(" u   Set time between each automatic bitrob event\n");
    DemoLogf(" v   Toggle verbose bitrob event reporting\n");
    DemoLogf(" V   Toggle verbose HDLC event reporting\n");
    DemoLogf(" j/J Start/Stop Automatic simulation of HDLC signalling\n");
    DemoLogf(" S   Set Sleep time between each HDLC frame\n");
    DemoLogf(" L   Set length of HDLC frame to be transmitted\n");
#if 0 // for testing a full event queue condition
    DemoLogf("\ng_bSimulateBusy=%d\n", g_bSimulateBusy);
#endif
    break;

  case 'v':
    {
      g_bBitRobSignallingVerbose = !g_bBitRobSignallingVerbose;
      DemoLogf("g_bBitRobSignallingVerbose is now %s.\n", g_bBitRobSignallingVerbose ? "TRUE": "FALSE");
    }
    break;

  case 'V':
    {
      g_bHdlcSignallingVerbose = !g_bHdlcSignallingVerbose;
      DemoLogf("g_bHdlcSignallingVerbose is now %s.\n", g_bHdlcSignallingVerbose ? "TRUE": "FALSE");
    }
    break;

    case 'z':
      {
        DemoLogf("Setting transmitted BR data (all channels) to 0 on LI#%d \n", nCurrentLi);
        if (g_demoEnv.bChannelizedAbcdBitAccess) {
          OTX_INT32 i;
          for (i = 0; i < 24; i++) {
            result = OtxT1SetChannelizedRobBitData(g_hLogicDevs.hT1E1Cas[nCurrentLi], i+1, 0); DemoCheckRc(result);
          }
        } else {
          memset(&g_sqSendRobBitData[nCurrentLi], 0, sizeof(OtxT1RobBitDataS));
          result = OtxT1SetRobBitData(g_hLogicDevs.hT1E1Cas[nCurrentLi], &g_sqSendRobBitData[nCurrentLi]); DemoCheckRc(result);
        }
        DemoLogf("Done\n");
      }
      break;

    case 'Z':
      {
        OTX_INT32 nLi;
        
        DemoLogf("Setting transmitted BR data (all channels) to 0 on both LIs\n");
        
        for (nLi = 0; nLi < 2; nLi++) {
          
          if (g_demoEnv.bChannelizedAbcdBitAccess) {
            OTX_INT32 i;
            for (i = 0; i < 24; i++) {
              result = OtxT1SetChannelizedRobBitData(g_hLogicDevs.hT1E1Cas[nLi], i+1, 0); DemoCheckRc(result);
            }
          } else {
            memset(&g_sqSendRobBitData[nLi], 0, sizeof(OtxT1RobBitDataS));
            result = OtxT1SetRobBitData(g_hLogicDevs.hT1E1Cas[nLi], &g_sqSendRobBitData[nLi]); DemoCheckRc(result);
          }
        }
        DemoLogf("Done\n");
      }
      break;

    case 's':
      {
        char szABCDbuf[4+1];  // +1 for the '\0'
        char szBuf[12];  // +1 for the '\0'

        if (g_demoEnv.bChannelizedAbcdBitAccess) {
          OTX_UINT8 nABCDBits;

          DemoLogf("Enter ABCD-bit data (0x0 - 0xF) for CAS channel %d on Li%d: ", nCurrentCasChannel, nCurrentLi);
          OsaConsoleGetStr(szBuf);
          nABCDBits = (OTX_UINT8)strtoul(szBuf, NULL, 0);
          DemoLogf("Setting ABCD=0x%X for CAS channel %d on Li%d\n", nABCDBits, nCurrentCasChannel, nCurrentLi);
          result = OtxT1SetChannelizedRobBitData(g_hLogicDevs.hT1E1Cas[nCurrentLi], nCurrentCasChannel, nABCDBits); DemoCheckRc(result);

        } else {
          DemoGetCasChannelValue(nCurrentCasChannel, &g_sqSendRobBitData[nCurrentLi], szABCDbuf);
          
          DemoLogf("Enter ABCD-bit data for CAS channel %d on Li%d (ABCD=%s): ", nCurrentCasChannel, nCurrentLi, szABCDbuf);
          OsaConsoleGetStr(szBuf);
          if (*szBuf) {
            if (strlen(szBuf) != 4) {
              DemoLogf("Invalid entry (%s). Please enter ABCD-bit data as a 4 character string.\n", szBuf);
              DemoLogf("E.g. \"0000\", or \"0101\" \n");
              break;
            } else {
              DemoLogf("Setting ABCD=\"%s\" for CAS channel %d on Li%d\n", szBuf, nCurrentCasChannel, nCurrentLi);
              DemoSetCasChannelValue(nCurrentCasChannel, &g_sqSendRobBitData[nCurrentLi], szBuf);
              result = OtxT1SetRobBitData(g_hLogicDevs.hT1E1Cas[nCurrentLi], &g_sqSendRobBitData[nCurrentLi]); DemoCheckRc(result);
            }
          } else {
            DemoLogf("ABCD-bit data remains unchanged for this channel\n");
          }
          DemoLogf("Done.\n");
        }
      }
      break;

    case 'r':
      {
        OTX_INT32 nLi = nCurrentLi + 1;
        
        if ((nCurrentLi % 2) == 0) {
          // Even Li
          nLi = nCurrentLi + 1;
        } else {
          // Odd Li
          nLi = nCurrentLi - 1;
        }
        DemoPrintAllCasChannelsData(nLi);
      }
      break;

    case 'R':
      {
        DemoPrintAllCasChannelsData(nCurrentLi);
      }
      break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 i;

      for (i = 0; i < g_demoEnv.coLi; i++) {
        DemoLogf("Line status of Li #%d is ", i);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[i], &eLayer1Status); DemoCheckRc(result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;


  case 't':
    {
      static OTX_UINT8 nFrameCount = 0;
      OTX_INT32 i;
      OTX_INT32 busyCount = 0;

      if (g_cbHdlcFrame > OD_HDLC_MAX_TX_HDLC_FRAME_LEN) {
        DemoLogf("Error: Frame Length too long\n");
        break;
      }

      for (i = 0; i < g_cbHdlcFrame; i++) {
        g_sqHdlcFrame[i] = (OTX_UINT8) i;
      }
      g_sqHdlcFrame[0] = ++nFrameCount;

      DemoLogf("Sending %d HDLC frame(s) from Li #%d (len=%d, nFrameCount=0x%X) ...\n", g_coHdlcSenderFrames, nCurrentLi, g_cbHdlcFrame, nFrameCount);

      // Send HDLC Frame(s)
      for (i = 0; i < g_coHdlcSenderFrames; i++) {
        result = OtxHdlcSendData(g_hLogicDevs.hHdlcSender[nCurrentLi], g_sqHdlcFrame, g_cbHdlcFrame, DEMO_TASK_REF, OTX_TIME_INFINITY); 
        if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
          // Driver IO Request buffer is full. We need to retransmit this frame.
          i--;
          busyCount++;
        } else {
          DemoCheckRc(result);
          OsaThreadSleep(g_nSleepTimeBetweenHdlcFrames);
        }
      }

      DemoLogf("Done (BusyCount=%d).\n", busyCount);
    }
    break;


  case 'L':
    {
      char strBuf[12];
      DemoLogf("Enter HDLC Frame Length to be transmitted (%d) ", g_cbHdlcFrame);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_cbHdlcFrame = (OTX_INT32)strtol(strBuf, NULL, 0);
        DemoLogf("HDLC Frame Length is now set to %d\n", g_cbHdlcFrame);
      } else {
        DemoLogf("HDLC Frame Length remains unchanged\n");
      }
    }
    break;

  case 'S':
    {
      char strBuf[12];
      DemoLogf("Enter Time to Sleep (in ms) between transmitted HDLC Frames (%d) ", g_nSleepTimeBetweenHdlcFrames);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_nSleepTimeBetweenHdlcFrames = (OTX_INT32)strtol(strBuf, NULL, 0);
        DemoLogf("Sleep time is now set to %d ms\n", g_nSleepTimeBetweenHdlcFrames);
      } else {
        DemoLogf("Sleep time remains unchanged\n");
      }
    }
    break;

  case '+':
    {
      char strBuf[12];

      DemoLogf("Set the number of consecutive frame to sent when 't' are pressed (%d)\n", g_coHdlcSenderFrames);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_coHdlcSenderFrames = (OTX_INT32) strtol(strBuf, NULL, 0);
        DemoLogf("Consecutive frame count is now %d\n", g_coHdlcSenderFrames);
      } else {
        DemoLogf("Consecutive frame count remains unchanged\n");
      }
    }
    break;

  case 'y':
    {
      DemoLogf("Starting Simulation of Bitrob signalling\n");
      g_bSimulateBitRobSignalling = OTX_TRUE;
      OsaThreadStart(DemoSimulateBitRobSignalling);
      break;
    }

  case 'Y':
    {
      DemoLogf("Aborting Simulation of Bitrob signalling\n");
      g_bSimulateBitRobSignalling = OTX_FALSE;
      break;
    }

  case 'j':
    {
      DemoLogf("Starting Simulation of HDLC signalling\n");
      g_bSimulateHdlcSignalling = OTX_TRUE;
      OsaThreadStart(DemoSimulateHdlcSignalling);
      break;
    }

  case 'J':
    {
      DemoLogf("Aborting Simulation of HDLC signalling\n");
      g_bSimulateHdlcSignalling = OTX_FALSE;
      break;
    }

  case 'u':
    {
      char strBuf[12];
      DemoLogf("Enter Time to Sleep (in ms) between each Bitrob signalling event (%d) ", g_nSleepTimeBetweenBitrobEvents);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_nSleepTimeBetweenBitrobEvents = (OTX_INT32)strtol(strBuf, NULL, 0);
        DemoLogf("Sleep time is now set to %d ms\n", g_nSleepTimeBetweenBitrobEvents);
      } else {
        DemoLogf("Sleep time remains unchanged\n");
      }
    }
    break;



  case 'n':
    {
      // Select current nCurrentCasChannel
      nCurrentCasChannel++;
      if (nCurrentCasChannel > 24) {
        // Wrap 
        nCurrentCasChannel = 1;
      }
      
      DemoLogf("CAS Channel #%d Selected \n", nCurrentCasChannel);
      
    }
    break;
    
    
  case 'N':
    {
      // Select current nCurrentCasChannel
      nCurrentCasChannel--;
      if (nCurrentCasChannel < 1) {
        // Wrap
        nCurrentCasChannel = 24;
      }
      
      DemoLogf("CAS Channel #%d Selected \n", nCurrentCasChannel);
    }
    break;
    
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
    {
      // Select current Li
      OTX_INT32 nLi = commandChar - '0';

      if (nLi > g_demoEnv.coLi) {
        DemoLogf("Error: LI#%d is not supported by this board\n", nLi);
      } else {
        nCurrentLi = nLi;
        DemoLogf("LI#%d Selected ....\n", nCurrentLi);
      }
    }
    break;

  case 'i':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(result);

      DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'o':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, nValue); DemoCheckRc(result);

      DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}




/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);
}

/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;
  OTX_INT32 i;

	OsaConsoleInit();

  DemoLogf("T2pT1RobBit, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Example program for accessing T1 Rob-bit data\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);
  DemoLogf("ABCD Bit access is %s\n", g_demoEnv.bChannelizedAbcdBitAccess ? "channelized" : "non-channelized");

  for (i = 0; i < g_demoEnv.coLi; i++) {
    memset(&g_sqSendRobBitData[i], 0, sizeof(OtxT1RobBitDataS));
  }

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Set clock source for the board (important to set the clock source before configuring the E1/T1 devices)
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(result);

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoInitializeLogicalDevices();

  // Enable all devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnableDevices();


  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300);
  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}


#if defined(_WIN32_WCE)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd) 
{
  // default command line parameters
  WCHAR wsExeName[OTX_MAX_PATH * sizeof(WCHAR)];
  static char exename[OTX_MAX_PATH];
  static char *cmdParams[] = { exename, "/TM", NULL }; 
  char *pc;

  GetModuleFileName(NULL, wsExeName, sizeof(wsExeName));
  WideCharToMultiByte(CP_ACP, 0, &wsExeName[0], sizeof(wsExeName), exename, sizeof(exename), NULL, NULL); 
  OsaStrCpy(g_demoEnv.cwd, cmdParams[0]);
  pc = OsaStrRChr(g_demoEnv.cwd, '\\');
  if (pc != NULL) {
    *++pc = '\0';
  }

  return OsaWinCeConsoleMain(hInstance, hPrevInstance, lpCmdLine, nShowCmd,
              cmdParams, sizeof(cmdParams) / sizeof(cmdParams[0]), main);
}
#endif  // #if defined(_WIN32_WCE)
